package net.sf.sido.factory.json;

import static net.sf.sido.factory.json.JSONUtils.getBoolean;
import static net.sf.sido.factory.json.JSONUtils.getString;

import java.io.IOException;
import java.net.URL;

import net.sf.sido.DOJsonFieldNotArrayException;
import net.sf.sido.DOTypeNotFoundException;
import net.sf.sido.DataFactory;
import net.sf.sido.DataObject;
import net.sf.sido.DataSchema;
import net.sf.sido.DataType;
import net.sf.sido.DataTypeBuilder;
import net.sf.sido.array.Array;
import net.sf.sido.factory.DataFactoryReader;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.codehaus.jackson.JsonNode;
import org.codehaus.jackson.JsonParser;
import org.codehaus.jackson.map.ObjectMapper;

import com.google.common.base.Predicate;

public class JSONDataFactoryReader implements DataFactoryReader, JSONDataFactory {

	@Override
	public DataType getDOType(JSONReaderContext context, String type) throws IOException {
		// Reference to another schema
		if (type.contains(DataSchema.QUALIFIER_SEPARATOR)) {
			String schemaName = StringUtils.substringBefore(type, DataSchema.QUALIFIER_SEPARATOR);
			// Another schema
			if (!StringUtils.equals(context.getSchema().getName(), schemaName)) {
				DataType anotherType = context.getSchema().getFactory().getType(type);
				if (anotherType != null) {
					return anotherType;
				} else {
					throw new DOTypeNotFoundException(type);
				}
			}
		}
		// Gets the definition for this type
		JSONTypeDefinition tDefinition = context.getDefinition(type);
		if (tDefinition == null) {
			throw new DOJsonTypeDefinitionNotFoundException(type);
		}
		// If not resolved, resolve it
		JSONResolutionStatus status = tDefinition.getStatus();
		if (status == JSONResolutionStatus.COMPLETE) {
			return tDefinition.getType();
		} else if (status == JSONResolutionStatus.PENDING) {
			resolve(context, tDefinition);
			return tDefinition.getType();
		} else if (status == JSONResolutionStatus.ONGOING) {
			return tDefinition.getBuilder().getType();
		} else {
			throw new IllegalStateException(String.format("Status %s is unknown.", status));
		}
	}

	@Override
	public DataSchema read(URL url, DataFactory factory) throws IOException {
		// Checks
		Validate.notNull(url, "The schema URL is required");
		Validate.notNull(factory, "The schema factory is required");
		// Factory
		ObjectMapper mapper = new ObjectMapper();
		mapper.configure(JsonParser.Feature.ALLOW_UNQUOTED_FIELD_NAMES, true);
		// Reads
		JsonNode jRoot = mapper.readValue(url, JsonNode.class);

		// Gets the name
		String name = getString(jRoot, "name", true, null);

		// Gets the types
		JsonNode jTypes = jRoot.get("types");
		if (jTypes == null) {
			throw new DOSchemaTypesFieldNotFoundException(name, url);
		}

		// Creates the schema
		DataSchema schema = factory.createSchema(name);

		// Context
		JSONReaderContext context = new JSONReaderContext(schema);

		// Reads all types
		readTypes(context, jTypes);

		// Reads all extensions
		readExtensions(context, jRoot);

		// Closes the schema
		schema.close();

		// OK
		return schema;
	}

	protected void readExtension(DataType parentType, JSONReaderContext context, JsonNode jExtension) {
		// Properties
		String name = getString(jExtension, "name", true, null);
		String type = getString(jExtension, "type", false, "string");
		// Additional properties
		boolean nullable = getBoolean(jExtension, "nullable", false, false);
		boolean reverse = getBoolean(jExtension, "reverse", false, true);
		// Constraint
		Predicate<?> constraint = null;
		JsonNode jConstraint = jExtension.get("constraint");
		if (jConstraint != null && !jConstraint.isNull()) {
			// Makes the constraint using the factory
			constraint = context.getFactory().getConstraintFactory().createConstraint(
					new JSONConstraintDefinitions(jConstraint));
		}
		// Extends the type
		if (type.startsWith("*#")) {
			// DO type name
			type = type.substring(2);
			// Gets the type in the current schema
			DataType propertyType = context.getSchema().getType(type);
			if (propertyType == null) {
				throw new DOSchemaTypeNotFoundException(type, context.getSchema().getName());
			}
			@SuppressWarnings("unchecked")
			Predicate<Array<DataObject>> doConstraint = (Predicate<Array<DataObject>>) constraint;
			parentType.extendArray(name, propertyType, nullable, doConstraint);
		} else if (type.startsWith("*")) {
			type = type.substring(1);
			Class<Object> simpleType = context.getFactory().getSimpleType(type);
			@SuppressWarnings("unchecked")
			Predicate<Array<Object>> simpleConstraint = (Predicate<Array<Object>>) constraint;
			parentType.extendArray(name, simpleType, nullable, simpleConstraint);
		} else if (type.startsWith("#")) {
			// DO type name
			type = type.substring(1);
			// Gets the type in the current schema
			DataType propertyType = context.getSchema().getType(type);
			if (propertyType == null) {
				throw new DOSchemaTypeNotFoundException(type, context.getSchema().getName());
			}
			@SuppressWarnings("unchecked")
			Predicate<DataObject> doConstraint = (Predicate<DataObject>) constraint;
			parentType.extend(name, propertyType, nullable, reverse, doConstraint);
		} else {
			Class<Object> simpleType = context.getFactory().getSimpleType(type);
			@SuppressWarnings("unchecked")
			Predicate<Object> simpleConstraint = (Predicate<Object>) constraint;
			parentType.extend(name, simpleType, nullable, simpleConstraint);
		}
	}

	protected void readExtension(JSONReaderContext context, JsonNode jExtension) {
		// Gets the type to extend
		String parentName = getString(jExtension, "parent", true, null);
		DataType parentType = context.getSchema().getFactory().getType(parentName);
		if (parentType == null) {
			throw new DOTypeNotFoundException(parentName);
		}

		// 1 - simple declaration
		readExtension(parentType, context, jExtension);

		// 2 - list of extensions
		JsonNode jProperties = jExtension.get("properties");
		for(JsonNode jProperty : jProperties) {
			readExtension(parentType, context, jProperty);
		}
	}

	protected void readExtensions(JSONReaderContext context, JsonNode jRoot) {
		// Gets the extensions
		JsonNode jExtensions = jRoot.get("extensions");
		if (jExtensions != null) {
			// Checks this is an array
			if (!jExtensions.isArray()) {
				throw new DOJsonFieldNotArrayException("extensions");
			}
			// For each extension
			for(JsonNode jExtension : jExtensions) {
				readExtension(context, jExtension);
			}
		}
	}

	protected void readType(JSONReaderContext context, JsonNode jType) {
		// Gets the type name
		String name = getString(jType, "name", true, null);
		// Creates the type builder
		DataTypeBuilder builder = context.getSchema().createDataTypeBuilder(name);
		// Extensibility
		if (JSONUtils.getBoolean(jType, "extensible", false, false)) {
			builder.allowExtensions();
		}
		// Abstract type
		if (JSONUtils.getBoolean(jType, "abstract", false, false)) {
			builder.abstractType();
		}
		// Registers the builder together with the properties
		context.addDefinition(new JSONTypeDefinition(name, builder, jType));
	}

	protected void readTypes(JSONReaderContext context, JsonNode jTypes) throws IOException {
		// Checks this is an array
		if (!jTypes.isArray()) {
			throw new DOJsonFieldNotArrayException("types");
		}
		// Count of types
		int count = jTypes.size();
		for(int i = 0; i < count; i++) {
			// Type definition
			JsonNode jType = jTypes.get(i);
			// Reads the type
			readType(context, jType);
		}

		// All type definitions have now been read

		// Resolve the types and their properties
		resolve(context);
	}

	protected void resolve(JSONReaderContext context) throws IOException {
		// For all types
		for(JSONTypeDefinition definition : context.getDefinitions()) {
			JSONResolutionStatus status = definition.getStatus();
			if (status == JSONResolutionStatus.PENDING) {
				resolve(context, definition);
			} else if (status != JSONResolutionStatus.COMPLETE) {
				throw new IllegalStateException(String.format("Status %s for definition %s cannot be resolved.",
						status, definition.getName()));
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.sido.io.json.JSONDataFactory#resolve(net.sf.sido.io .json.JSONReaderContext,
	 * net.sf.sido.io.json.JSONTypeDefinition)
	 */
	@Override
	public void resolve(JSONReaderContext context, JSONTypeDefinition definition) throws IOException {
		definition.start();
		// Super-type
		String parentTypeName = JSONUtils.getString(definition.getNode(), "parent", false, null);
		if (StringUtils.isNotBlank(parentTypeName)) {
			// Gets the parent type reference
			DataType parentType = getDOType(context, parentTypeName);
			// Sets as parent
			definition.setParentType(parentType);
		}
		// Gets all properties
		JsonNode jProperties = definition.getNode().get("properties");
		if (jProperties != null) {
			if (jProperties.isArray()) {
				for(int i = 0; i < jProperties.size(); i++) {
					JsonNode jProperty = jProperties.get(i);
					resolveProperty(context, definition, jProperty);
				}
			} else {
				throw new DOJsonFieldNotArrayException("properties");
			}
		}
		// Completion
		definition.complete();
	}

	protected <T> void resolveProperty(JSONReaderContext context, JSONTypeDefinition definition, JsonNode jProperty)
			throws IOException {
		// Property name
		String name = JSONUtils.getString(jProperty, "name", true, null);
		// Type information
		String type = JSONUtils.getString(jProperty, "type", false, "string");
		// Creates the property assembler
		PropertyAssembler assembler;
		if ("#".equals(type)) {
			// Anonymous type
			assembler = new AnonymousPropertyAssembler(name);
		} else if ("*#".equals(type)) {
			// Array of generics
			assembler = new AnonymousArrayPropertyAssembler(name);
		} else if (type.startsWith("*#")) {
			assembler = new ArrayDOPropertyAssembler(name, type.substring(2));
		} else if (type.startsWith("*")) {
			assembler = new ArraySimplePropertyAssembler(name, type.substring(1));
		} else if (type.startsWith("#")) {
			assembler = new DOPropertyAssembler(name, type.substring(1));
		} else {
			assembler = new SimplePropertyAssembler(name, type);
		}
		// Creates the property
		assembler.addProperty(context, this, definition, jProperty);
	}
}
