package net.sf.sido.factory.sido.antlr;

import java.util.ArrayList;
import java.util.List;

import net.sf.sido.DataFactory;
import net.sf.sido.DataSchema;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SidoReporter implements ISidoReporter {

	private static final Logger log = LoggerFactory.getLogger(SidoReporter.class);

	private final DataFactory factory;

	private DataSchema schema;

	private IPropertyContainer container;
	private List<IPropertyContainer> containers = new ArrayList<IPropertyContainer>();

	public SidoReporter(DataFactory factory) {
		this.factory = factory;
	}

	@Override
	public void close() {
		log.debug("close");
		// Resolve all containers
		for (IPropertyContainer propertyContainer : containers) {
			propertyContainer.resolve();
		}
		// Closes all types
		schema.close();
	}

	@Override
	public void closeContainer() {
		log.debug("container close");
		containers.add(container);
		container = null;
	}

	public DataSchema getSchema() {
		if (schema == null) {
			throw new IllegalStateException("No schema has been parsed.");
		} else {
			return schema;
		}
	}

	@Override
	public void pushAbstract() {
		log.debug("abstract");
		if (container instanceof TypePropertyContainer) {
			((TypePropertyContainer) container).setAbstract();
		} else {
			throw new IllegalStateException("No type is available.");
		}
	}

	@Override
	public void pushAnonymousProperty() {
		log.debug("anonymous");
		container.pushAnonymousProperty();
	}

	@Override
	public void pushConstraint(String id) {
		log.debug("constraint " + id);
		container.pushConstraint(id);
	}

	@Override
	public void pushConstraintParam(String name, String value) {
		log.debug("constraint param " + name + " = " + value);
		container.pushConstraintParam(name, value);
	}

	@Override
	public void pushConstraintValue(String value) {
		log.debug("constraint value " + value);
		container.pushConstraintValue(value);
	}

	@Override
	public void pushExtensible() {
		log.debug("extensible");
		if (container instanceof TypePropertyContainer) {
			((TypePropertyContainer) container).setExtensible();
		} else {
			throw new IllegalStateException("No type is available.");
		}
	}

	@Override
	public void pushExtension(String forType) {
		log.debug("extension for " + forType);
		container = new ExtensionPropertyContainer(schema, forType);
	}

	@Override
	public void pushNewProperty() {
		log.debug("property new");
		container.pushNewProperty();
	}

	@Override
	public void pushParent(String parentTypeName) {
		log.debug("parent " + parentTypeName);
		if (container instanceof TypePropertyContainer) {
			((TypePropertyContainer) container).setParentTypeName(parentTypeName);
		} else {
			throw new IllegalStateException("No type is available.");
		}
	}

	@Override
	public void pushProperty(String name) {
		name = StringUtils.stripStart(name, "#");
		log.debug("property " + name);
		container.pushProperty(name);
	}

	@Override
	public void pushPropertyArray() {
		log.debug("array ");
		container.pushPropertyArray();
	}

	@Override
	public void pushPropertyNullable(boolean not) {
		log.debug("nullable " + not);
		container.pushPropertyNullable(not);
	}

	@Override
	public void pushPropertyReverse() {
		log.debug("reverse");
		container.pushPropertyReverse();
	}

	@Override
	public void pushPropertyType(String typeName) {
		log.debug("type " + typeName);
		container.pushPropertyType(typeName);
	}

	@Override
	public void pushSchemaName(String name) {
		log.debug("schema " + name);
		if (schema != null) {
			throw new IllegalStateException("Schema has already been declared");
		} else {
			schema = factory.createSchema(name);
		}
	}

	@Override
	public void pushTypeName(String name) {
		log.debug("type " + name);
		container = new TypePropertyContainer(schema, name);
	}

}
