package net.sf.sido.io;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;

import net.sf.sido.DONullPropertyException;
import net.sf.sido.DOTypeNotFoundException;
import net.sf.sido.DataFactory;
import net.sf.sido.DataObject;
import net.sf.sido.DataType;
import net.sf.sido.Property;
import net.sf.sido.PropertyMetaInfos;
import net.sf.sido.array.Array;
import net.sf.sido.editor.DataEditor;
import net.sf.sido.support.DOCache;

import org.apache.commons.lang.StringUtils;

/**
 * Default implementation for a {@link DataObjectReader} that relies on reading some attributes and elements from the
 * input, in a format-independent way.
 * @author Damien Coraboeuf
 * @param <T> Type of element to read from (XML element, JSON node, etc.)
 */
public abstract class AbstractDataObjectReader<T> implements DataObjectReader<T> {

	protected abstract int getDOId(T e);

	protected abstract int getDOReference(T e);

	protected abstract String getDOType(T e, boolean required);

	protected abstract T getElement(T e, Property<?> property);

	protected abstract Iterable<T> getElements(T e, Property<?> property);

	protected abstract String getText(T eNode);

	protected DataObject read(DOCache cache, T e, DataFactory factory, DataType expectedType) {
		DataType type;
		// No expected type
		if (expectedType == null) {
			// Gets the mandatory type name
			String typeName = getDOType(e, true);
			// Gets the type from factory
			type = factory.getType(typeName);
			if (type == null) {
				throw new DOTypeNotFoundException(typeName);
			}
		} else {
			// Gets any additional type name
			String typeName = getDOType(e, false);
			// Checks if defined
			if (StringUtils.isNotBlank(typeName)) {
				// Gets the type from factory
				type = factory.getType(typeName);
				if (type == null) {
					throw new DOTypeNotFoundException(typeName);
				}
				// Asserts the tow types are identical
				if (!type.equals(expectedType)) {
					throw new DOExpectedTypeException(expectedType, type);
				}
			}
			// OK
			type = expectedType;
		}
		// Reference ?
		int ref = getDOReference(e);
		if (ref > 0) {
			// Gets the object from the cache
			DataObject o = cache.get(type, ref);
			if (o != null) {
				return o;
			} else {
				throw new DOMissingReferenceException(type.getName(), ref);
			}
		}
		// Gets the ID
		int id = getDOId(e);
		// Creates the instance
		DataObject o = type.newInstance();
		// Registers the instance
		/*
		 * Note that the ID of the instance may be different that the one which is read.
		 */
		if (id == 0) {
			id = o.getSidoId();
		}
		cache.put(id, o);
		// For all properties
		for(Property<?> property : type.getProperties()) {
			read(cache, e, o, property, factory);
		}
		// OK
		return o;
	}

	protected <P> void read(DOCache cache, T e, DataObject o, Property<P> property, DataFactory factory) {
		// Property name
		String name = property.getName();
		// Data object property?
		boolean doProperty = property.getMetaInfo().getBoolean(PropertyMetaInfos.DO, false);
		// Array property ?
		if (property.isArray()) {
			// Collection of items
			Collection<P> items = new ArrayList<P>();
			// Gets the array elements
			Iterable<T> eArray = getElements(e, property);
			if (eArray != null) {
				// // For each child
				for(T eElement : eArray) {
					// Data object?
					if (doProperty) {
						// Item property
						DataType propertyType = property.getMetaInfo().getObject(PropertyMetaInfos.ELEMENT_TYPE, null);
						@SuppressWarnings("unchecked")
						P child = (P) read(cache, eElement, factory, propertyType);
						items.add(child);
					}
					// Simple property
					else {
						@SuppressWarnings("unchecked")
						P value = (P) simple(eElement, property.getMetaInfo().getType(PropertyMetaInfos.ELEMENT_TYPE),
								factory);
						items.add(value);
					}
				}
				// Creates the array
				Array<P> array = new Array<P>(items);
				// Sets the array
				o.setArray(name, array);
			}
		}
		// Single data object property
		else if (doProperty) {
			// Item property
			DataType propertyType = property.getMetaInfo().getObject(PropertyMetaInfos.DO_TYPE, null);
			// Reads the node
			T eChild = getElement(e, property);
			if (eChild != null) {
				// Transforms the node
				DataObject child = read(cache, eChild, factory, propertyType);
				// Sets the property
				o.setValue(name, child);
			}
		}
		// Simple property
		else {
			// Gets the field value
			T eNode = getElement(e, property);
			if (eNode != null) {
				Object value = simple(eNode, property.getType(), factory);
				o.setValue(name, value);
			} else if (!property.isNullable()) {
				throw new DONullPropertyException(o, name);
			}
		}
	}

	@Override
	public DataObject read(InputStream in, DataFactory factory) throws IOException {
		return read(in, factory, null);
	}

	@Override
	public DataObject read(T element, DataFactory factory, DataType expectedType) {
		return read(new DOCache(), element, factory, expectedType);
	}

	@Override
	public DataObject read(T element, DataFactory factory) {
		return read(element, factory, null);
	}

	protected Object simple(T eNode, Class<?> type, DataFactory factory) {
		// Value
		String value = getText(eNode);
		// Gets the editor
		DataEditor<?> editor = factory.getEditor(type);
		// Object value
		Object t = editor.valueOf(value);
		// Output
		return t;
	}

}
