package net.sf.sido.io.xml;

import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

import net.sf.sido.DataObject;
import net.sf.sido.DataType;
import net.sf.sido.Property;
import net.sf.sido.io.AbstractDataObjectWriter;
import net.sf.sido.io.DOReferenceManagement;

import org.dom4j.Document;
import org.dom4j.DocumentFactory;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.XMLWriter;

public class XMLDataObjectWriter extends AbstractDataObjectWriter<Element, QName> {

	private final boolean prettyFormat;
	private final XMLDataObjectNamespace doNamespace;

	private final DocumentFactory eFactory = DocumentFactory.getInstance();

	public XMLDataObjectWriter() {
		this(false);
	}

	public XMLDataObjectWriter(boolean prettyFormat) {
		this(prettyFormat, new DefaultXMLDataObjectNamespace());
	}

	public XMLDataObjectWriter(boolean prettyFormat, XMLDataObjectNamespace dataObjectNamespace) {
		this.prettyFormat = prettyFormat;
		doNamespace = dataObjectNamespace;
	}

	public XMLDataObjectWriter(XMLDataObjectNamespace dataObjectNamespace) {
		this(false, dataObjectNamespace);
	}

	@Override
	protected void add(Element e, QName propertyName, Element propertyValue) {
		e.add(propertyValue);
	}

	@Override
	protected void createArray(Element e, QName pName, List<Element> elements) {
		Element eArray = createElement(pName);
		e.add(eArray);
		for (Element eChild : elements) {
			eArray.add(eChild);
		}
	}

	@Override
	protected Element createElement(QName name) {
		return eFactory.createElement(name);
	}

	@Override
	protected Element createValue(DataObject o, QName pName, Object value) {
		// Gets the value as a string
		String text = getText(o.getFactory(), value);
		// Adds the value
		if (text != null) {
			Element eChild = eFactory.createElement(pName);
			eChild.setText(text);
			return eChild;
		} else {
			return null;
		}
	}

	protected void declareNamespace(Element e, Namespace namespace) {
		e = getRoot(e);
		e.add(namespace);
	}

	@Override
	protected QName getDOArrayName() {
		return XMLDOSchema.ELE_DATAOBJECT;
	}

	@Override
	protected QName getPropertyName(Element e, Property<?> property) {
		// Qualified name for the property
		QName name = doNamespace.getQName(property);
		// Declares, if needed, the namespace at root level
		declareNamespace(e, name.getNamespace());
		// OK
		return name;
	}

	protected Element getRoot(Element e) {
		Element p = e.getParent();
		if (p != null) {
			return getRoot(p);
		} else {
			return e;
		}
	}

	@Override
	protected QName getSimpleArrayName() {
		return XMLDOSchema.ELE_VALUE;
	}

	@Override
	protected void setDOId(Element e, int id) {
		e.addAttribute(XMLDOSchema.ATT_ID, String.valueOf(id));
	}

	@Override
	protected void setDOReference(Element e, int id) {
		e.addAttribute(XMLDOSchema.ATT_REF, String.valueOf(id));
	}

	@Override
	protected void setDOType(Element e, DataType type) {
		e.addAttribute(XMLDOSchema.ATT_TYPE, type.toString());
	}

	@Override
	public void write(OutputStream out, DataObject o, DOReferenceManagement referenceManagement) throws IOException {
		// Creates a DOM
		Document dom = eFactory.createDocument();
		// Creates the element for the data object
		Element eData = createElement(XMLDOSchema.ELE_DATAOBJECT, o, referenceManagement);
		// Appends the element to the DOM
		dom.setRootElement(eData);
		// Configuration of the output (pretty or compact)
		OutputFormat outputFormat = prettyFormat ? OutputFormat.createPrettyPrint() : OutputFormat.createCompactFormat();
		// Writes this DOM
		new XMLWriter(out, outputFormat).write(dom);

	}
}
